@@ -1,13 +1,13 @@
 /***************************************
-  $Header: /home/amb/procmeter3/modules/RCS/acpi.c 1.9 2004/12/11 16:11:32 amb Exp $
+  $Header: /home/amb/CVS/procmeter3/modules/acpi.c,v 1.17 2010-02-28 10:22:11 amb Exp $
 
-  ProcMeter - A system monitoring program for Linux - Version 3.4d.
+  ProcMeter - A system monitoring program for Linux - Version 3.5d.
 
   ACPI source file.
   ******************/ /******************
   Written by Joey Hess.
 
-  This file Copyright 2001 Joey Hess
+  This file Copyright 2001-2007 Joey Hess
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -28,7 +28,7 @@
 #include <dirent.h>
 #include <string.h>
 #ifdef ACPI_APM
-#include <apm.h>
+#include "apm.h"
 #endif
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -87,6 +87,7 @@ char *acpi_labels_20020214[] = {
 	"thermal_zone",
 #endif
 	"state:",
+	"last full capacity:",
 	NULL
 };
 
@@ -107,7 +108,7 @@ inline char *get_acpi_file (const char *file) {
 	fd = open(file, O_RDONLY);
 	if (fd == -1) return NULL;
 	end = read(fd, buf, sizeof(buf));
-	buf[end] = '\0';
+	buf[end-1] = '\0';
 	close(fd);
 	return buf;
 }
@@ -131,7 +132,7 @@ inline char *scan_acpi_value (const char *buf, const char *key) {
 	static char ret[256];
 	
 	if ((ptr = strstr(buf, key))) {
-		if (sscanf(ptr + strlen(key), "%s", ret) == 1)
+		if (sscanf(ptr + strlen(key), "%255s", ret) == 1)
 			return ret;
 	}
 	return NULL;
@@ -147,18 +148,37 @@ char *get_acpi_value (const char *file, const char *key) {
 	return scan_acpi_value(buf, key);
 }
 
-/* Returns the design capacity of a battery. */
+/* Returns the maximum capacity of a battery.
+ *
+ * Note that this returns the highest possible capacity for the battery,
+ * even if it can no longer charge that fully. So normally it uses the
+ * design capacity. While the last full capacity of the battery should
+ * never exceed the design capacity, some silly hardware might report
+ * that it does. So if the last full capacity is greater, it will be
+ * returned.
+ */
 int get_acpi_batt_capacity(int battery) {
-	int cap;
-	char *caps=get_acpi_value(acpi_batt_info[battery], acpi_labels[label_design_capacity]);
-	if (caps == NULL)
-		cap=0; /* battery not present */
+	int dcap, lcap;
+	char *s;
+
+	s=get_acpi_value(acpi_batt_info[battery], acpi_labels[label_design_capacity]);
+	if (s == NULL)
+		dcap=0; /* battery not present */
 	else
-		cap=atoi(caps);
+		dcap=atoi(s);
+
 	/* This is ACPI's broken way of saying that there is no battery. */
-	if (cap == 655350)
+	if (dcap == 655350)
 		return 0;
-	return cap;
+
+	s=get_acpi_value(acpi_batt_info[battery], acpi_labels[label_last_full_capacity]);
+	if (s != NULL) {
+		lcap=atoi(s);
+		if (lcap > dcap)
+			return lcap;
+	}
+
+	return dcap;
 }
 
 /* Comparison function for qsort. */
@@ -187,8 +207,8 @@ int find_items (char *itemname, char infoarray[ACPI_MAXITEM][128],
 	if (dir == NULL)
 		return 0;
 	while ((ent = readdir(dir))) {
-		if (!strncmp(".", ent->d_name, 1) || 
-		    !strncmp("..", ent->d_name, 2))
+		if (!strcmp(".", ent->d_name) || 
+		    !strcmp("..", ent->d_name))
 			continue;
 
 		devices[num_devices]=strdup(ent->d_name);
@@ -217,7 +237,6 @@ int find_items (char *itemname, char infoarray[ACPI_MAXITEM][128],
 int find_batteries(void) {
 	int i;
 	acpi_batt_count = find_items(acpi_labels[label_battery], acpi_batt_info, acpi_batt_status);
-	/* Read in the last charged capacity of the batteries. */
 	for (i = 0; i < acpi_batt_count; i++)
 		acpi_batt_capacity[i] = get_acpi_batt_capacity(i);
 	return acpi_batt_count;
@@ -264,17 +283,23 @@ int acpi_supported (void) {
 	}
 	closedir(dir);
 	
-	version = get_acpi_value("info", "ACPI-CA Version:");
+	/* If kernel is 2.6.21 or newer, version is in
+	   /sys/module/acpi/parameters/acpica_version */
+	
+	version = get_acpi_file("/sys/module/acpi/parameters/acpica_version");
 	if (version == NULL) {
-		/* 2.5 kernel acpi */
-		version = get_acpi_value("info", "version:");
+		version = get_acpi_value(PROC_ACPI "/info", "ACPI-CA Version:");
+		if (version == NULL) {
+			/* 2.5 kernel acpi */
+			version = get_acpi_value(PROC_ACPI "/info", "version:");
+		}
 	}
 	if (version == NULL) {
 		return 0;
 	}
 	num = atoi(version);
 	if (num < ACPI_VERSION) {
-		fprintf(stderr, "ACPI subsystem %s too is old, consider upgrading to %i.\n",
+		fprintf(stderr, "ProcMeter(%s): ACPI subsystem %s too is old, consider upgrading to %i.\n",__FILE__,
 				version, ACPI_VERSION);
 		return 0;
 	}
@@ -300,13 +325,22 @@ int acpi_supported (void) {
 int acpi_read (int battery, apm_info *info) {
 	char *buf, *state;
 	
+	if (acpi_batt_count == 0) {
+		info->battery_percentage = 0;
+		info->battery_time = 0;
+		info->battery_status = BATTERY_STATUS_ABSENT;
+		acpi_batt_capacity[battery] = 0;
+		/* Where else would the power come from, eh? ;-) */
+		info->ac_line_status = 1;
+		return 0;
+	}
+	
 	/* Internally it's zero indexed. */
 	battery--;
 	
 	buf = get_acpi_file(acpi_batt_status[battery]);
 	if (buf == NULL) {
-		fprintf(stderr, "unable to read %s\n", acpi_batt_status[battery]);
-		perror("read");
+                fprintf(stderr, "ProcMeter(%s): Unable to read %s\n",__FILE__, acpi_batt_status[battery]);
 		exit(1);
 	}
 
@@ -342,12 +376,19 @@ int acpi_read (int battery, apm_info *info) {
 		if (state) {
 			if (state[0] == 'd') { /* discharging */
 				info->battery_status = BATTERY_STATUS_CHARGING;
+				/* Expensive ac power check used here
+				 * because AC power might be on even if a
+				 * battery is discharging in some cases. */
+				info->ac_line_status = on_ac_power();
 			}
 			else if (state[0] == 'c' && state[1] == 'h') { /* charging */
 				info->battery_status = BATTERY_STATUS_CHARGING;
 				info->ac_line_status = 1;
 				info->battery_flags = info->battery_flags | BATTERY_FLAGS_CHARGING;
-				info->battery_time = -1 * (float) (acpi_batt_capacity[battery] - pcap) / (float) rate * 60;
+				if (rate)
+					info->battery_time = -1 * (float) (acpi_batt_capacity[battery] - pcap) / (float) rate * 60;
+				else
+					info->battery_time = 0;
 				if (abs(info->battery_time) < 0.5)
 					info->battery_time = 0;
 			}
@@ -358,6 +399,10 @@ int acpi_read (int battery, apm_info *info) {
 			}
 			else if (state[0] == 'c') { /* not charging, so must be critical */
 				info->battery_status = BATTERY_STATUS_CRITICAL;
+				/* Expensive ac power check used here
+				 * because AC power might be on even if a
+				 * battery is critical in some cases. */
+				info->ac_line_status = on_ac_power();
 			}
 			else if (rate == 0) {
 				/* if rate is null, battery charged, on
@@ -366,7 +411,7 @@ int acpi_read (int battery, apm_info *info) {
 				info->ac_line_status = 1;
 			}
 			else {
-				fprintf(stderr, "unknown battery state: %s\n", state);
+                                fprintf(stderr, "ProcMeter(%s): Unknown battery state: %s\n", __FILE__, state);
 			}
 		}
 		else {
@@ -382,8 +427,13 @@ int acpi_read (int battery, apm_info *info) {
 			 * buf below this point! */
 			acpi_batt_capacity[battery] = get_acpi_batt_capacity(battery);
 		}
+		else if (pcap > acpi_batt_capacity[battery]) {
+			/* Battery is somehow charged to greater than max
+			 * capacity. Rescan for a new max capacity. */
+			find_batteries();
+		}
 		
-		if (pcap) {
+		if (pcap && acpi_batt_capacity[battery]) {
 			/* percentage = (current_capacity / max capacity) * 100 */
 			info->battery_percentage = (float) pcap / (float) acpi_batt_capacity[battery] * 100;
 		}
@@ -397,7 +447,7 @@ int acpi_read (int battery, apm_info *info) {
 		info->battery_time = 0;
 		info->battery_status = BATTERY_STATUS_ABSENT;
 		acpi_batt_capacity[battery] = 0;
-		if (acpi_batt_count == 1) {
+		if (acpi_batt_count == 0) {
 			/* Where else would the power come from, eh? ;-) */
 			info->ac_line_status = 1;
 		}
@@ -467,13 +517,13 @@ ProcMeterOutput _batt_outputs[N_BATT_OUTPUTS]=
  /*+ Rate of charge/discharge. +*/
  {
   /* char  name[];          */ "Batt%i_Rate",
-  /* char *description;     */ "The present rate of battery %i discharge or charge, in mW.",
+  /* char *description;     */ "The present rate of battery %i discharge or charge, in mA.",
   /* char  type;            */ PROCMETER_GRAPH|PROCMETER_TEXT|PROCMETER_BAR,
   /* short interval;        */ 1,
-  /* char  text_value[];    */ "0 mW",
+  /* char  text_value[];    */ "0 mA",
   /* long  graph_value;     */ 0,
   /* short graph_scale;     */ 1000,
-  /* char  graph_Units[];   */ "(%d mW)"
+  /* char  graph_Units[];   */ "(%d mA)"
  },
  /*+ The battery status output +*/
  {
@@ -745,7 +795,7 @@ NOBATT:
 	
 		/* Charge/discharge rate. */
 		batt_outputs[index + 1].graph_value = PROCMETER_GRAPH_FLOATING((float)rate/batt_outputs[index + 1].graph_scale);
-		sprintf(batt_outputs[index + 1].text_value,"%i mW", rate);
+		sprintf(batt_outputs[index + 1].text_value,"%i mA", rate);
 	}
 
 	return(0);
